@@ -9,13 +9,15 @@ use Devel::Declare ();
 use Parse::Method::Signatures;
 use Moose::Meta::Class;
 use Moose::Util::TypeConstraints;
-use MooseX::Meta::Signature::Combined;
-use MooseX::Types::Moose qw/Str Any/;
+use Moose::Util qw/does_role/;
+use MooseX::Types::Moose qw/Str Defined Maybe Object ArrayRef/;
+use MooseX::Types::Structured qw/Dict Tuple Optional/;
+use aliased 'Parse::Method::Signatures::Param::Named';
 use MooseX::Method::Signatures::Meta::Method;
 
 use namespace::clean -except => 'meta';
 
-our $VERSION = '0.08';
+our $VERSION = '0.09';
 
 extends qw/Moose::Object Devel::Declare::MethodInstaller::Simple/;
 
@@ -46,7 +48,7 @@ sub setup_for {
 sub param_to_spec {
     my ($self, $param) = @_;
 
-    my $tc = Any;
+    my $tc = Defined;
     $tc = $param->meta_type_constraint
         if $param->has_type_constraints;
 
@@ -57,9 +59,14 @@ sub param_to_spec {
     }
 
     my %spec;
-    $spec{ required } = $param->required ? 1 : 0;
-    $spec{ isa      } = $tc if defined $tc;
-    $spec{ default  } = $param->default_value if $param->has_default_value;
+    $spec{tc} = $param->required
+        ? $tc
+        : does_role($param, Named)
+            ? Optional[$tc]
+            : Maybe[$tc];
+
+    $spec{default} = $param->default_value
+        if $param->has_default_value;
 
     if ($param->has_traits) {
         for my $trait (@{ $param->param_traits }) {
@@ -76,7 +83,7 @@ sub parse_proto {
     $proto ||= '';
 
     my $vars = q{};
-    my @param_spec;
+    my (@named, @positional);
 
     my $sig = Parse::Method::Signatures->signature(
         input => "(${proto})",
@@ -94,28 +101,74 @@ sub parse_proto {
     if ($sig->has_invocant) {
         my $invocant = $sig->invocant;
         $vars .= $invocant->variable_name . q{,};
-        push @param_spec, $self->param_to_spec($invocant);
+        push @positional, $self->param_to_spec($invocant);
     }
     else {
         $vars .= '$self,';
-        push @param_spec, { required => 1 };
+        push @positional, { tc => Object };
     }
 
     if ($sig->has_positional_params) {
         for my $param ($sig->positional_params) {
             $vars .= $param->variable_name . q{,};
-            push @param_spec, $self->param_to_spec($param);
+            push @positional, $self->param_to_spec($param);
         }
     }
 
     if ($sig->has_named_params) {
         for my $param ($sig->named_params) {
             $vars .= $param->variable_name . q{,};
-            push @param_spec, $param->label => $self->param_to_spec($param);
+            push @named, $param->label => $self->param_to_spec($param);
         }
     }
 
-    return ($sig, $vars, \@param_spec);
+    my $tc = Tuple[
+        Tuple[ map { $_->{tc}               } @positional ],
+        Dict[  map { ref $_ ? $_->{tc} : $_ } @named      ],
+    ];
+
+    my $coerce_param = sub {
+        my ($spec, $value) = @_;
+        return $value unless exists $spec->{coerce};
+        return $spec->{tc}->coerce($value);
+    };
+
+    my %named = @named;
+
+    coerce $tc,
+        from ArrayRef,
+        via {
+            my (@positional_args, %named_args);
+
+            my $i = 0;
+            for my $param (@positional) {
+                push @positional_args,
+                    $#{ $_ } < $i
+                        ? (exists $param->{default} ? $param->{default} : ())
+                        : $coerce_param->($param, $_->[$i]);
+                $i++;
+            }
+
+            unless ($#{ $_ } < $i) {
+                my %rest = @{ $_ }[$i .. $#{ $_ }];
+                while (my ($key, $spec) = each %named) {
+                    if (exists $rest{$key}) {
+                        $named_args{$key} = $coerce_param->($spec, delete $rest{$key});
+                        next;
+                    }
+
+                    if (exists $spec->{default}) {
+                        $named_args{$key} = $spec->{default};
+                    }
+                }
+
+                @named_args{keys %rest} = values %rest;
+            }
+
+            return [\@positional_args, \%named_args];
+        };
+
+    return ($sig, $vars, [@positional, @named], $tc);
 }
 
 sub inject_parsed_proto {
@@ -131,7 +184,7 @@ sub parser {
     my $name   = $self->strip_name;
     my $proto  = $self->strip_proto;
     my $attrs  = $self->strip_attrs;
-    my ($sig, $vars, $param_spec) = $self->parse_proto($proto);
+    my ($sig, $vars, $param_spec, $tc) = $self->parse_proto($proto);
     my $inject = $self->inject_parsed_proto($vars);
 
     if (defined $name) {
@@ -155,11 +208,12 @@ sub parser {
     my $create_meta_method = sub {
         my ($code) = @_;
         return MooseX::Method::Signatures::Meta::Method->wrap(
-            _signature   => $sig,
-            _param_spec  => $param_spec,
-            body         => $code,
-            package_name => $pkg,
-            name         => $meth_name,
+            _signature      => $sig,
+            _param_spec     => $param_spec,
+            body            => $code,
+            package_name    => $pkg,
+            name            => $meth_name,
+            type_constraint => $tc,
         );
     };
 
